import pysplashsurf
import numpy as np


def sphere_sdf_mc_test(dtype):
    radius = 1.0
    num_verts = 100

    grid_size = radius * 2.2
    dx = grid_size / (num_verts - 1)

    translation = -0.5 * grid_size

    def make_sdf():
        coords = np.arange(num_verts, dtype=dtype) * dx + translation
        x, y, z = np.meshgrid(coords, coords, coords, indexing="ij")
        sdf = np.sqrt(x**2 + y**2 + z**2) - radius
        return sdf

    sdf = make_sdf()

    # Note: Currently this reconstruction assumes that inside the surface values get bigger (like a density function)
    mesh, grid = pysplashsurf.marching_cubes(
        sdf, iso_surface_threshold=0.0, cube_size=dx, translation=[translation] * 3, return_grid=True
    )

    assert len(mesh.vertices) > 0

    norms = np.linalg.norm(mesh.vertices, axis=1)
    assert norms.min() > radius - 1e-4
    assert norms.max() < radius + 1e-4

    assert pysplashsurf.check_mesh_consistency(mesh, grid) is None


def test_sphere_sdf_mc_f32():
    sphere_sdf_mc_test(np.float32)


def test_sphere_sdf_mc_f64():
    sphere_sdf_mc_test(np.float64)



import pysplashsurf
import numpy as np
import meshio
import subprocess
import time
import trimesh
import pathlib

BINARY_PATH = "splashsurf"
DIR = pathlib.Path(__file__).parent.resolve()
VTK_PATH = DIR.joinpath("ParticleData_Fluid_5.vtk")


def now_s():
    return time.process_time_ns() / (10**9)


def marching_cubes_calls(dtype):
    print("\nTesting marching cubes calls")

    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)
    reconstruction = pysplashsurf.reconstruct_surface(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0,
        cube_size=0.5,
        iso_surface_threshold=0.6,
    )
    mesh = reconstruction.mesh
    verts_before = len(mesh.vertices)
    print("# of vertices before:", verts_before)

    mesh_with_data = pysplashsurf.MeshWithData(mesh)
    pysplashsurf.marching_cubes_cleanup(mesh_with_data, reconstruction.grid)

    mesh = mesh_with_data.mesh
    verts_after = len(mesh.vertices)
    print("# of vertices after:", verts_after)
    assert verts_after < verts_before


def test_marching_cubes_calls_f32():
    marching_cubes_calls(np.float32)


def test_marching_cubes_calls_f64():
    marching_cubes_calls(np.float64)


def reconstruction_pipeline(
    input_file,
    output_file,
    *,
    attributes_to_interpolate=None,
    multi_threading=True,
    particle_radius=0.025,
    rest_density=1000.0,
    smoothing_length=2.0,
    cube_size=0.5,
    iso_surface_threshold=0.6,
    mesh_smoothing_weights=False,
    output_mesh_smoothing_weights=False,
    sph_normals=False,
    mesh_smoothing_weights_normalization=13.0,
    mesh_smoothing_iters=5,
    normals_smoothing_iters=5,
    mesh_aabb_min=None,
    mesh_aabb_max=None,
    mesh_cleanup=False,
    decimate_barnacles=False,
    keep_vertices=False,
    compute_normals=False,
    output_raw_normals=False,
    output_raw_mesh=False,
    mesh_aabb_clamp_vertices=False,
    check_mesh_closed=False,
    check_mesh_manifold=False,
    check_mesh_orientation=False,
    check_mesh_debug=False,
    generate_quads=False,
    quad_max_edge_diag_ratio=1.75,
    quad_max_normal_angle=10.0,
    quad_max_interior_angle=135.0,
    subdomain_grid=False,
    subdomain_num_cubes_per_dim=64,
    dtype
):
    mesh = meshio.read(input_file)
    particles = np.array(mesh.points, dtype=dtype)

    if attributes_to_interpolate is None:
        attributes_to_interpolate = []

    # Prepare attributes dictionary
    attrs = {}
    for attr in attributes_to_interpolate:
        if attr in mesh.point_data:
            if mesh.point_data[attr].dtype.kind == "f":
                attrs[attr] = mesh.point_data[attr].astype(dtype)
            else:
                attrs[attr] = mesh.point_data[attr].astype(np.int64)

    mesh_with_data, reconstruction = pysplashsurf.reconstruction_pipeline(
        particles,
        attributes_to_interpolate=attrs,
        multi_threading=multi_threading,
        particle_radius=particle_radius,
        rest_density=rest_density,
        smoothing_length=smoothing_length,
        cube_size=cube_size,
        iso_surface_threshold=iso_surface_threshold,
        mesh_smoothing_weights=mesh_smoothing_weights,
        sph_normals=sph_normals,
        mesh_smoothing_weights_normalization=mesh_smoothing_weights_normalization,
        mesh_smoothing_iters=mesh_smoothing_iters,
        normals_smoothing_iters=normals_smoothing_iters,
        mesh_aabb_min=mesh_aabb_min,
        mesh_aabb_max=mesh_aabb_max,
        mesh_cleanup=mesh_cleanup,
        decimate_barnacles=decimate_barnacles,
        keep_vertices=keep_vertices,
        compute_normals=compute_normals,
        output_raw_normals=output_raw_normals,
        output_raw_mesh=output_raw_mesh,
        mesh_aabb_clamp_vertices=mesh_aabb_clamp_vertices,
        subdomain_grid=subdomain_grid,
        subdomain_num_cubes_per_dim=subdomain_num_cubes_per_dim,
        output_mesh_smoothing_weights=output_mesh_smoothing_weights,
        check_mesh_closed=check_mesh_closed,
        check_mesh_manifold=check_mesh_manifold,
        check_mesh_orientation=check_mesh_orientation,
        check_mesh_debug=check_mesh_debug,
        generate_quads=generate_quads,
        quad_max_edge_diag_ratio=quad_max_edge_diag_ratio,
        quad_max_normal_angle=quad_max_normal_angle,
        quad_max_interior_angle=quad_max_interior_angle,
    )

    mesh_with_data.write_to_file(output_file)


def no_post_processing_test(dtype):
    start = now_s()
    subprocess.run(
        [BINARY_PATH]
        + f"reconstruct {VTK_PATH} -o {DIR.joinpath('test_bin.vtk')} -r=0.025 -l=2.0 -c=0.5 -t=0.6 {"-d=on" if dtype == np.float64 else ""} --subdomain-grid=on --mesh-cleanup=off --mesh-smoothing-weights=off --mesh-smoothing-iters=0 --normals=off --normals-smoothing-iters=0".split(),
        check=True,
    )
    print("Binary done in", now_s() - start)

    start = now_s()
    reconstruction_pipeline(
        VTK_PATH,
        DIR.joinpath("test.vtk"),
        particle_radius=0.025,
        smoothing_length=2.0,
        cube_size=0.5,
        iso_surface_threshold=0.6,
        mesh_smoothing_weights=False,
        mesh_smoothing_iters=0,
        normals_smoothing_iters=0,
        mesh_cleanup=False,
        compute_normals=False,
        subdomain_grid=True,
        dtype=dtype
    )
    print("Python done in", now_s() - start)

    binary_mesh = meshio.read(DIR.joinpath("test_bin.vtk"))
    python_mesh = meshio.read(DIR.joinpath("test.vtk"))

    binary_verts = np.array(binary_mesh.points, dtype=dtype)
    python_verts = np.array(python_mesh.points, dtype=dtype)

    print("# of vertices binary:", len(binary_verts))
    print("# of vertices python:", len(python_verts))

    assert len(binary_verts) == len(python_verts)

    binary_verts.sort(axis=0)
    python_verts.sort(axis=0)

    assert np.allclose(binary_verts, python_verts)


def test_no_post_processing_f32():
    no_post_processing_test(np.float32)


def test_no_post_processing_f64():
    no_post_processing_test(np.float64)


def with_post_processing_test(dtype):
    start = now_s()
    subprocess.run(
        [BINARY_PATH]
        + f"reconstruct {VTK_PATH} -o {DIR.joinpath('test_bin.vtk')} -r=0.025 -l=2.0 -c=0.5 -t=0.6 {"-d=on" if dtype == np.float64 else ""} --subdomain-grid=on --interpolate_attribute velocity --decimate-barnacles=on --mesh-cleanup=on --mesh-smoothing-weights=on --mesh-smoothing-iters=25 --normals=on --normals-smoothing-iters=10 --output-smoothing-weights=on --generate-quads=off".split(),
        check=True,
    )
    print("Binary done in", now_s() - start)

    start = now_s()
    reconstruction_pipeline(
        VTK_PATH,
        DIR.joinpath("test.vtk"),
        attributes_to_interpolate=["velocity"],
        particle_radius=0.025,
        smoothing_length=2.0,
        cube_size=0.5,
        iso_surface_threshold=0.6,
        mesh_smoothing_weights=True,
        mesh_smoothing_weights_normalization=13.0,
        mesh_smoothing_iters=25,
        normals_smoothing_iters=10,
        generate_quads=False,
        mesh_cleanup=True,
        compute_normals=True,
        subdomain_grid=True,
        decimate_barnacles=True,
        output_mesh_smoothing_weights=True,
        output_raw_normals=True,
        dtype=dtype
    )
    print("Python done in", now_s() - start)

    binary_mesh = meshio.read(DIR.joinpath("test_bin.vtk"))
    python_mesh = meshio.read(DIR.joinpath("test.vtk"))

    # Compare number of vertices
    binary_verts = np.array(binary_mesh.points, dtype=dtype)
    python_verts = np.array(python_mesh.points, dtype=dtype)

    print("# of vertices binary:", len(binary_verts))
    print("# of vertices python:", len(python_verts))

    assert len(binary_verts) == len(python_verts)

    # Compare interpolated attribute
    binary_vels = binary_mesh.point_data["velocity"]
    python_vels = python_mesh.point_data["velocity"]

    binary_vels.sort(axis=0)
    python_vels.sort(axis=0)

    assert np.allclose(binary_vels, python_vels)

    # Trimesh similarity test
    # TODO: Replace load_mesh call: the function tries to create temporary files which may fail on some CI runners
    binary_mesh = trimesh.load_mesh(DIR.joinpath("test_bin.vtk"), "vtk")
    python_mesh = trimesh.load_mesh(DIR.joinpath("test.vtk"), "vtk")

    (_, distance_bin, _) = trimesh.proximity.closest_point(binary_mesh, python_verts)
    (_, distance_py, _) = trimesh.proximity.closest_point(python_mesh, binary_verts)
    distance = (np.sum(distance_bin) + np.sum(distance_py)) / (
        len(distance_bin) + len(python_verts)
    )
    print("Distance:", distance)
    assert distance < 1e-5

    # Naïve similarity test

    binary_verts.sort(axis=0)
    python_verts.sort(axis=0)

    print("Binary verts:", binary_verts)
    print("Python verts:", python_verts)

    assert np.allclose(binary_verts, python_verts)


def test_with_post_processing_f32():
    with_post_processing_test(np.float32)


def test_with_post_processing_f64():
    with_post_processing_test(np.float64)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import sys

from sphinx.ext.autodoc.importer import import_module

# Verify that the generated stub file exists
stub_path = os.path.abspath("../../pysplashsurf.pyi")
if not os.path.exists(stub_path):
   print(f"### Error: Stub file not found at {stub_path}. Please ensure stub_gen produced pysplashsurf.pyi.")
   sys.exit(1)

# Import the stub file
print(f"### Prepend to path: {os.path.abspath('../../')}")
sys.path.insert(0, os.path.abspath("../../"))
pysplashsurf = import_module("pysplashsurf")

#import pysplashsurf

# -- Project information -----------------------------------------------------

project = "pySplashsurf"
copyright = "2025, Interactive Computer Graphics"
author = "Interactive Computer Graphics"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "numpydoc",
    "myst_parser",
    "sphinx_rtd_theme",
    #'sphinx_autodoc_typehints'
]

source_suffix = [".rst", ".md"]

numpydoc_class_members_toctree = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

autodoc_typehints = "both"

always_document_param_types = True
always_use_bars_union = True

napoleon_use_rtype = False
napoleon_include_special_with_doc = True

# typehints_document_rtype = False
# typehints_use_rtype = False
# typehints_use_signature = True
# typehints_use_signature_return = True



import pysplashsurf
import numpy as np
import meshio
import pathlib

DIR = pathlib.Path(__file__).parent.resolve()
BGEO_PATH = DIR.joinpath("ParticleData_Fluid_50.bgeo")


def bgeo_test(dtype):
    particles = np.array(meshio.read(BGEO_PATH).points, dtype=dtype)

    assert len(particles) == 4732


def test_bgeo_f32():
    bgeo_test(np.float32)


def test_bgeo_f64():
    bgeo_test(np.float64)



# The contents of this file are partially copied from the blender sequence loader (https://github.com/InteractiveComputerGraphics/blender-sequence-loader)

try:
    import gzip
    import numpy as np
    import meshio


    def readbgeo_to_meshio(filepath):
        with gzip.open(filepath, 'r') as file:
            byte = file.read(5)
            if byte != b"BgeoV":
                raise Exception('not bgeo file format')
            byte = file.read(4)
            version = int.from_bytes(byte, byteorder="big")
            if version != 5:
                raise Exception('bgeo file not version 5')

            header = {}
            point_attributes = {}
            point_attributes_names = []
            point_attributes_sizes = []
            point_attributes_types = []
            position = None

            byte = file.read(4)
            header['nPoints'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nPrims'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nPointGroups'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nPrimGroups'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nPointAttrib'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nVertexAttrib'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nPrimAttrib'] = int.from_bytes(byte, byteorder="big")

            byte = file.read(4)
            header['nAttrib'] = int.from_bytes(byte, byteorder="big")

            particle_size = 4

            for _ in range(header['nPointAttrib']):
                byte = file.read(2)
                namelength = int.from_bytes(byte, byteorder="big")
                name_binary = file.read(namelength)
                name = name_binary.decode('utf-8')
                point_attributes_names.append(name)

                byte = file.read(2)
                size = int.from_bytes(byte, byteorder="big")
                point_attributes_sizes.append(size)
                particle_size += size

                byte = file.read(4)
                input_dtype = int.from_bytes(byte, byteorder="big")
                if input_dtype == 0:
                    point_attributes_types.append('FLOAT')
                    # read default value
                    # not going to do anything about it
                    byte = file.read(size * 4)
                elif input_dtype == 1:
                    point_attributes_types.append('INT')
                    # read default value
                    # not going to do anything about it
                    byte = file.read(size * 4)
                elif input_dtype == 5:
                    point_attributes_types.append('VECTOR')
                    # read default value
                    # not going to do anything about it
                    byte = file.read(size * 4)
                else:
                    raise Exception('input_dtype unknown/unsupported')
            byte = file.read(particle_size * header['nPoints'] * 4)
            # > means big endian
            attribute_data = np.frombuffer(byte, dtype='>f')
            attribute_data = np.reshape(attribute_data, (header['nPoints'], particle_size))
            # the first 3 column is its position data
            position = attribute_data[:, :3]
            # the 4th column is homogeneous coordiante, which is all 1, and will be ignored

            current_attribute_start_point = 4
            for i in range(header['nPointAttrib']):
                if point_attributes_types[i] == 'FLOAT':
                    point_attributes[point_attributes_names[i]] = attribute_data[:, current_attribute_start_point]
                    current_attribute_start_point += 1
                elif point_attributes_types[i] == 'VECTOR':
                    point_attributes[
                        point_attributes_names[i]] = attribute_data[:,
                                                                    current_attribute_start_point:current_attribute_start_point + 3]
                    current_attribute_start_point += 3
                elif point_attributes_types[i] == 'INT':
                    data = (attribute_data[:, current_attribute_start_point]).tobytes()
                    # > means big endian
                    point_attributes[point_attributes_names[i]] = np.frombuffer(data, dtype='>i')
                    current_attribute_start_point += 1
            remaining = file.read()
            if not remaining == b'\x00\xff':
                raise Exception("file didn't end")
            return meshio.Mesh(position, [('vertex', [])], point_data=point_attributes)


    # no need for write function
    meshio.register_format("bgeo", [".bgeo"], readbgeo_to_meshio, {".bgeo": None})

except ImportError:
    pass


# This file is automatically generated by pyo3_stub_gen
# ruff: noqa: E501, F401

from __future__ import annotations

import builtins
import numpy
import numpy.typing
import os
import pathlib
import typing
from enum import Enum

class Aabb3d:
    r"""
    Three-dimensional axis-aligned bounding box defined by its minimum and maximum corners
    """
    @property
    def min(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        The min coordinate of the AABB
        """
    @property
    def max(self) -> numpy.typing.NDArray[numpy.float64]:
        r"""
        The max coordinate of the AABB
        """
    @staticmethod
    def from_min_max(min:typing.Sequence[builtins.float], max:typing.Sequence[builtins.float]) -> Aabb3d:
        r"""
        Constructs an AABB with the given min and max coordinates
        """
    @staticmethod
    def from_points(points:numpy.typing.NDArray[typing.Any]) -> Aabb3d:
        r"""
        Constructs the smallest AABB fitting around all the given points
        """
    def contains_point(self, point:typing.Sequence[builtins.float]) -> builtins.bool:
        r"""
        Checks if the given point is inside the AABB, the AABB is considered to be half-open to its max coordinate
        """

class MeshAttribute:
    @property
    def dtype(self) -> numpy.dtype:
        r"""
        Numpy dtype of the data stored in the attribute
        """
    @property
    def name(self) -> builtins.str:
        r"""
        Name of the attribute
        """
    @property
    def data(self) -> numpy.typing.NDArray[typing.Any]:
        r"""
        View of the attribute data as a numpy array
        """

class MeshWithData:
    r"""
    Mesh with attached point and cell attributes
    """
    @property
    def dtype(self) -> numpy.dtype:
        r"""
        Numpy dtype of the underlying scalar type (either ``np.float32`` or ``np.float64``)
        """
    @property
    def nvertices(self) -> builtins.int:
        r"""
        Number of vertices in the mesh
        """
    @property
    def ncells(self) -> builtins.int:
        r"""
        Number of cells (triangles or quads) in the mesh
        """
    @property
    def mesh_type(self) -> MeshType:
        r"""
        Type of the underlying mesh
        """
    @property
    def mesh(self) -> typing.Union[TriMesh3d, MixedTriQuadMesh3d]:
        r"""
        The wrapped mesh without associated data and attributes
        """
    @property
    def point_attributes(self) -> dict[str, numpy.typing.NDArray]:
        r"""
        The attributes attached points (vertices) of the mesh
        """
    @property
    def cell_attributes(self) -> dict[str, numpy.typing.NDArray]:
        r"""
        The attributes attached to the cells (triangles or quads) of the mesh
        """
    def __new__(cls, mesh:typing.Union[TriMesh3d, MeshWithData]) -> MeshWithData:
        r"""
        Wraps an existing mesh object (either `TriMesh3d` or `MixedTriQuadMesh3d`) such that data (point and cell attributes) can be attached to it
        """
    def copy_mesh(self) -> typing.Union[TriMesh3d, MixedTriQuadMesh3d]:
        r"""
        Returns a copy of the wrapped mesh without associated data and attributes
        """
    def copy(self) -> MeshWithData:
        r"""
        Returns a copy (deep copy) of this mesh with its data and attributes
        """
    def add_point_attribute(self, name:builtins.str, attribute:numpy.typing.NDArray[typing.Any]) -> None:
        r"""
        Attaches a point attribute to the mesh
        
        There has to be exactly one attribute value per vertex in the mesh.
        As attribute data, the following numpy array types are supported:
         - 1D array with shape (N,) of ``np.uint64``
         - 1D array with shape (N,) of the mesh scalar type (``np.float32`` or ``np.float64``)
         - 2D array with shape (N,3) of the mesh scalar type (``np.float32`` or ``np.float64``)
        The data is copied into the mesh object.
        """
    def add_cell_attribute(self, name:builtins.str, attribute:numpy.typing.NDArray[typing.Any]) -> None:
        r"""
        Attaches a cell attribute to the mesh
        
        There has to be exactly one attribute value per cell in the mesh.
        As attribute data, the following numpy array types are supported:
         - 1D array with shape (N,) of ``np.uint64``
         - 1D array with shape (N,) of the mesh scalar type (``np.float32`` or ``np.float64``)
         - 2D array with shape (N,3) of the mesh scalar type (``np.float32`` or ``np.float64``)
        The data is copied into the mesh object.
        """
    def write_to_file(self, path:builtins.str | os.PathLike | pathlib.Path, *, file_format:typing.Optional[builtins.str]='vtk42') -> None:
        r"""
        Writes the mesh and its attributes to a file using ``meshio.write_points_cells``
        """

class MixedTriQuadMesh3d:
    r"""
    Mixed triangle and quad surface mesh in 3D
    """
    @property
    def dtype(self) -> numpy.dtype:
        r"""
        Numpy dtype of the underlying scalar type (either ``np.float32`` or ``np.float64``)
        """
    @property
    def vertices(self) -> numpy.typing.NDArray[typing.Any]:
        r"""
        The `Nx3` array of vertex positions of the mesh
        """
    def copy(self) -> MixedTriQuadMesh3d:
        r"""
        Returns a copy (deep copy) of this mesh
        """
    def get_triangles(self) -> numpy.typing.NDArray[numpy.uint64]:
        r"""
        Returns a copy of all triangle cells of the mesh as an `Nx3` array of vertex indices
        """
    def get_quads(self) -> numpy.typing.NDArray[numpy.uint64]:
        r"""
        Returns a copy of all quad cells of the mesh as an `Nx4` array of vertex indices
        """
    def write_to_file(self, path:builtins.str | os.PathLike | pathlib.Path, *, file_format:typing.Optional[builtins.str]='vtk42') -> None:
        r"""
        Writes the mesh to a file using ``meshio.write_points_cells``
        """

class NeighborhoodLists:
    r"""
    Per particle neighborhood lists
    """
    def __len__(self) -> builtins.int:
        r"""
        Returns the number of particles for which neighborhood lists are stored
        """
    def __getitem__(self, idx:builtins.int) -> builtins.list[builtins.int]:
        r"""
        Returns the neighborhood list for the particle at the given index
        """
    def get_neighborhood_lists(self) -> builtins.list[builtins.list[builtins.int]]:
        r"""
        Returns all stored neighborhood lists as a list of lists
        """

class SphInterpolator:
    r"""
    Interpolator of per-particle quantities to arbitrary points using SPH interpolation (with cubic kernel)
    """
    def __new__(cls, particle_positions:numpy.typing.NDArray[typing.Any], particle_densities:numpy.typing.NDArray[typing.Any], particle_rest_mass:builtins.float, compact_support_radius:builtins.float) -> SphInterpolator:
        r"""
        Constructs an SPH interpolator (with cubic kernels) for the given particles
        
        Parameters
        ----------
        particle_positions : numpy.ndarray
            A two-dimensional numpy array of shape (N, 3) containing the positions of the particles that are used for interpolation.
        particle_densities : numpy.ndarray
            A one-dimensional numpy array of shape (N,) containing the densities of the particles.
        particle_rest_mass
            The rest mass of each particle (assumed to be the same for all particles).
        compact_support_radius
            The compact support radius of the cubic spline kernel used for interpolation.
        """
    def interpolate_quantity(self, particle_quantity:numpy.typing.NDArray[typing.Any], interpolation_points:numpy.typing.NDArray[typing.Any], *, first_order_correction:builtins.bool=False) -> numpy.typing.NDArray[typing.Any]:
        r"""
        Interpolates a scalar or vectorial per particle quantity to the given points
        """
    def interpolate_normals(self, interpolation_points:numpy.typing.NDArray[typing.Any]) -> numpy.typing.NDArray[typing.Any]:
        r"""
        Interpolates surface normals (i.e. normalized SPH gradient of the indicator function) of the fluid to the given points using SPH interpolation
        """

class SurfaceReconstruction:
    r"""
    Result returned by surface reconstruction functions with surface mesh and other data
    """
    @property
    def grid(self) -> UniformGrid:
        r"""
        The marching cubes grid parameters used for the surface reconstruction
        """
    @property
    def particle_densities(self) -> typing.Optional[numpy.typing.NDArray[typing.Any]]:
        r"""
        The global array of particle densities (`None` if they were only computed locally)
        """
    @property
    def particle_inside_aabb(self) -> typing.Optional[numpy.typing.NDArray[typing.Any]]:
        r"""
        A boolean array indicating whether each particle was inside the AABB used for the reconstruction (`None` if no AABB was set)
        """
    @property
    def particle_neighbors(self) -> typing.Optional[NeighborhoodLists]:
        r"""
        The global neighborhood lists per particle (`None` if they were only computed locally)
        """
    @property
    def mesh(self) -> TriMesh3d:
        r"""
        The reconstructed triangle mesh
        """

class TriMesh3d:
    r"""
    Triangle surface mesh in 3D
    """
    @property
    def dtype(self) -> numpy.dtype:
        r"""
        Numpy dtype of the underlying scalar type (either ``np.float32`` or ``np.float64``)
        """
    @property
    def vertices(self) -> numpy.typing.NDArray[typing.Any]:
        r"""
        The `Nx3` array of vertex positions of the mesh
        """
    @property
    def triangles(self) -> numpy.typing.NDArray[numpy.uint64]:
        r"""
        The `Mx3` array of vertex indices per triangle
        """
    def copy(self) -> TriMesh3d:
        r"""
        Returns a copy (deep copy) of this mesh
        """
    def vertex_normals_parallel(self) -> numpy.typing.NDArray[typing.Any]:
        r"""
        Computes the vertex normals of the mesh using an area weighted average of the adjacent triangle faces
        """
    def vertex_vertex_connectivity(self) -> VertexVertexConnectivity:
        r"""
        Computes the vertex-vertex connectivity of the mesh
        """
    def write_to_file(self, path:builtins.str | os.PathLike | pathlib.Path, *, file_format:typing.Optional[builtins.str]='vtk42') -> None:
        r"""
        Writes the mesh to a file using ``meshio.write_points_cells``
        """

class UniformGrid:
    r"""
    Struct containing the parameters of the uniform grid used for the surface reconstruction
    """
    @property
    def aabb(self) -> Aabb3d:
        r"""
        The AABB of the grid containing all marching cubes vertices influenced by the particle kernels
        """
    @property
    def cell_size(self) -> builtins.float:
        r"""
        Returns the cell size of the uniform grid (the marching cubes voxel size)
        """
    @property
    def npoints_per_dim(self) -> builtins.list[builtins.int]:
        r"""
        Returns the number of points (marching cubes vertices) per dimension in the uniform grid
        """
    @property
    def ncells_per_dim(self) -> builtins.list[builtins.int]:
        r"""
        Returns the number of cells (marching cubes voxels) per dimension in the uniform grid
        """

class VertexVertexConnectivity:
    r"""
    Vertex-vertex connectivity of a mesh
    """
    def copy_connectivity(self) -> builtins.list[builtins.list[builtins.int]]:
        r"""
        Returns a copy of the wrapped connectivity data
        """
    def take_connectivity(self) -> builtins.list[builtins.list[builtins.int]]:
        r"""
        Returns the wrapped connectivity data by moving it out of this object (zero copy)
        """

class MeshType(Enum):
    r"""
    Enum specifying the type of mesh wrapped by a ``MeshWithData``
    """
    Tri3d = ...
    r"""
    3D triangle mesh
    """
    MixedTriQuad3d = ...
    r"""
    3D mixed triangle and quad mesh
    """

def barnacle_decimation(mesh:typing.Union[TriMesh3d, MeshWithData], *, keep_vertices:builtins.bool) -> typing.Union[TriMesh3d, MeshWithData]:
    r"""
    Performs specialized decimation on the given mesh to prevent "barnacles" when applying weighted Laplacian smoothing
    
    The decimation is performed inplace and modifies the given mesh.
    Returns the vertex-vertex connectivity of the decimated mesh which can be used for other
    post-processing steps.
    
    Parameters
    ----------
    mesh
        The triangle mesh to decimate.
    keep_vertices
        Flag to retain any vertices without connectivity resulting from decimation instead of filtering them out.
    """

def check_mesh_consistency(mesh:typing.Union[TriMesh3d, MeshWithData], grid:UniformGrid, *, check_closed:builtins.bool=True, check_manifold:builtins.bool=True, debug:builtins.bool=False) -> typing.Optional[builtins.str]:
    r"""
    Checks the consistency of a reconstructed surface mesh (watertightness, manifoldness), optionally returns a string with details, if problems are found
    
    Parameters
    ----------
    mesh
        The triangle mesh to check for consistency.
    grid
        The uniform grid that was used for the marching cubes triangulation of the input mesh.
    check_closed
        Flag to enable checking if the mesh is closed (watertight).
    check_manifold
        Flag to enable checking if the mesh is manifold (i.e. has no non-manifold vertices & edges).
    debug
        Flag to enable additional debug output during the consistency checks.
    
    Returns
    -------
        An optional string with details about the problems found during the consistency checks.
        If no problems are found, None is returned.
    """

def convert_tris_to_quads(mesh:typing.Union[TriMesh3d, MeshWithData], *, non_squareness_limit:builtins.float=1.75, normal_angle_limit:builtins.float=10.0, max_interior_angle:builtins.float=135.0) -> typing.Union[MixedTriQuadMesh3d, MeshWithData]:
    r"""
    Converts triangles to quads by merging triangles sharing an edge if they fulfill the given criteria
    
    This operation creates a new mesh and does not modify the input mesh.
    Angles are specified in degrees.
    
    Parameters
    ----------
    mesh
        The triangle mesh to convert to a mixed triangle-quad mesh.
    non_squareness_limit
        Maximum allowed ratio of quad edge lengths to its diagonals to merge two triangles to a quad (inverse is used for minimum).
    normal_angle_limit
        Maximum allowed angle (in degrees) between triangle normals to merge them to a quad.
    max_interior_angle
        Maximum allowed vertex interior angle (in degrees) inside a quad to merge two triangles to a quad.
    """

def laplacian_smoothing_normals_parallel(normals:numpy.typing.NDArray[typing.Any], vertex_connectivity:VertexVertexConnectivity, *, iterations:builtins.int) -> None:
    r"""
    Laplacian smoothing of a normal field
    
    The smoothing is performed inplace and modifies the given normal array.
    
    Parameters
    ----------
    normals
        A two-dimensional array of shape (N, 3) containing the normals to smooth.
    vertex_connectivity
        The vertex-vertex connectivity of the mesh, required for efficient smoothing.
    iterations
        The number of smoothing iterations to perform.
    """

def laplacian_smoothing_parallel(mesh:typing.Union[TriMesh3d, MeshWithData], vertex_connectivity:VertexVertexConnectivity, *, iterations:builtins.int, beta:builtins.float=1.0, weights:numpy.typing.NDArray[typing.Any]) -> None:
    r"""
    Laplacian smoothing of mesh vertices with feature weights
    
    The smoothing is performed inplace and modifies the vertices of the given mesh.
    
    Parameters
    ----------
    mesh
        The triangle mesh to smooth.
    vertex_connectivity
        The vertex-vertex connectivity of the mesh, required for efficient smoothing.
    iterations
        The number of smoothing iterations to perform.
    beta
        Factor used for blending the original vertex position with the smoothed position.
    weights
        A one-dimensional array of weights per vertex that influence the smoothing.
        The weight is multiplied with beta.
    """

def marching_cubes(values:numpy.typing.NDArray[typing.Any], *, iso_surface_threshold:builtins.float, cube_size:builtins.float, translation:typing.Optional[typing.Sequence[builtins.float]]=None, return_grid:builtins.bool=False) -> typing.Union[TriMesh3d, tuple[TriMesh3d, UniformGrid]]:
    r"""
    Performs a standard marching cubes triangulation of a 3D array of values
    
    The array of values has to be a contiguous array with shape ``(nx, ny, nz)``.
    The iso-surface threshold defines which value is considered to be "on" the surface.
    The cube size and translation parameters define the scaling and translation of the resulting
    mesh. Without translation, the value ``values[0, 0, 0]`` is located at coordinates ``(0, 0, 0)``.
    
    The values are interpreted as a "density field", meaning that values higher than the iso-surface
    threshold are considered to be "inside" the surface and values lower than the threshold are
    considered to be "outside" the surface. This is the opposite convention to an SDF (signed distance field).
    However, even if values of an SDF are provided as an input, the marching cubes algorithm
    will still work and produce a watertight surface mesh (if the surface is fully contained in the
    array).
    
    If ``return_grid`` is set to ``True``, the function will return a tuple of the mesh and the
    uniform grid that was used for the triangulation. This can be used for other functions such as
    :py:func:`check_mesh_consistency`. Otherwise, only the mesh is returned.
    
    The function is currently single-threaded. The SPH surface reconstruction functions :py:func:`reconstruction_pipeline`
    and :py:func:`reconstruct_surface` improve performance by processing multiple patches in parallel.
    
    Parameters
    ----------
    values : numpy.ndarray
       A three-dimensional numpy array of shape (nx, ny, nz) containing the scalar values at the vertices
       of the marching cubes grid.
    iso_surface_threshold
       The iso-surface threshold value used to determine the surface.
    cube_size
       The size of each cube/voxel of the marching cubes grid. Determines the scaling of the resulting mesh.
    translation
       An optional translation vector [tx, ty, tz] applied to the entire mesh after scaling.
    """

def marching_cubes_cleanup(mesh:typing.Union[TriMesh3d, MeshWithData], grid:UniformGrid, *, max_rel_snap_dist:typing.Optional[builtins.float]=None, max_iter:builtins.int=5, keep_vertices:builtins.bool=False) -> typing.Union[TriMesh3d, MeshWithData]:
    r"""
    Performs simplification on the given mesh inspired by the "Compact Contouring"/"Mesh displacement" approach by Doug Moore and Joe Warren
    
    The simplification is performed inplace and modifies the given mesh.
    The method is designed specifically for meshes generated by Marching Cubes.
    See Moore and Warren: `Mesh Displacement: An Improved Contouring Method for Trivariate Data <https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.49.5214&rep=rep1&type=pdf>`_ (1991)
    or Moore and Warren: "Compact Isocontours from Sampled Data" in "Graphics Gems III" (1992).
    
    Parameters
    ----------
    mesh
        The triangle mesh to simplify.
    grid
        The uniform grid that was used for the marching cubes triangulation of the input mesh.
    max_rel_snap_dist
        Optional maximum relative snapping distance (relative to the grid cell size) to merge close vertices.
    max_iter
        The maximum number of iterations of cleanup to perform.
    keep_vertices
        Flag to retain any vertices without connectivity resulting from simplification instead of filtering them out.
    """

def neighborhood_search_spatial_hashing_parallel(particle_positions:numpy.typing.NDArray[typing.Any], domain:Aabb3d, search_radius:builtins.float) -> NeighborhoodLists:
    r"""
    Performs a neighborhood search using spatial hashing (multithreaded implementation)
    
    Parameters
    ----------
    particles : numpy.ndarray
        A two-dimensional numpy array of shape (N, 3) containing the positions of the particles.
    domain
        An axis-aligned bounding box (AABB) of the particles used for spatial hashing.
        The neighborhood search fails if particles are outside the domain.
    search_radius
        The radius per particle where other particles are considered neighbors.
    """

def reconstruct_surface(particles:numpy.typing.NDArray[typing.Any], *, particle_radius:builtins.float, rest_density:builtins.float=1000.0, smoothing_length:builtins.float, cube_size:builtins.float, iso_surface_threshold:builtins.float=0.6, aabb_min:typing.Optional[typing.Sequence[builtins.float]]=None, aabb_max:typing.Optional[typing.Sequence[builtins.float]]=None, multi_threading:builtins.bool=True, simd:builtins.bool=True, global_neighborhood_list:builtins.bool=False, subdomain_grid:builtins.bool=True, subdomain_grid_auto_disable:builtins.bool=True, subdomain_num_cubes_per_dim:builtins.int=64) -> SurfaceReconstruction:
    r"""
    Performs a surface reconstruction from the given particles without additional post-processing
    
    Note that all parameters use absolute distance units and are not relative to the particle radius.
    
    Parameters
    ----------
    particles : numpy.ndarray
        A two-dimensional numpy array of shape (N, 3) containing the positions of the particles.
    particle_radius
        Particle radius.
    rest_density
        Rest density of the fluid.
    smoothing_length
        Smoothing length of the SPH kernel in absolute distance units (compact support radius of SPH kernel will be twice the smoothing length).
    cube_size
        Size of the cubes (voxels) used for the marching cubes grid in absolute distance units.
    iso_surface_threshold
        Threshold of the SPH interpolation of the "color field" where the iso surface should be extracted.
    aabb_min
        Lower corner of the AABB of particles to consider in the reconstruction.
    aabb_max
        Upper corner of the AABB of particles to consider in the reconstruction.
    multi_threading
        Flag to enable multi-threading for the reconstruction and post-processing steps.
    simd
       Flag to enable SIMD vectorization for the reconstruction if supported by the CPU architecture.
    subdomain_grid
        Flag to enable spatial decomposition by dividing the domain into subdomains with dense marching cube grids for efficient multi-threading.
    subdomain_grid_auto_disable
        Flag to automatically disable the subdomain grid if the global domain is too small.
    subdomain_num_cubes_per_dim
        Number of marching cubes voxels along each coordinate axis in each subdomain if the subdomain grid is enabled.
    """

def reconstruction_pipeline(particles:numpy.typing.NDArray[typing.Any], *, attributes_to_interpolate:typing.Optional[dict]=None, particle_radius:builtins.float, rest_density:builtins.float=1000.0, smoothing_length:builtins.float, cube_size:builtins.float, iso_surface_threshold:builtins.float=0.6, aabb_min:typing.Optional[typing.Sequence[builtins.float]]=None, aabb_max:typing.Optional[typing.Sequence[builtins.float]]=None, multi_threading:builtins.bool=True, simd:builtins.bool=True, subdomain_grid:builtins.bool=True, subdomain_grid_auto_disable:builtins.bool=True, subdomain_num_cubes_per_dim:builtins.int=64, check_mesh_closed:builtins.bool=False, check_mesh_manifold:builtins.bool=False, check_mesh_orientation:builtins.bool=False, check_mesh_debug:builtins.bool=False, mesh_cleanup:builtins.bool=False, mesh_cleanup_snap_dist:typing.Optional[builtins.float]=None, decimate_barnacles:builtins.bool=False, keep_vertices:builtins.bool=False, compute_normals:builtins.bool=False, sph_normals:builtins.bool=False, normals_smoothing_iters:typing.Optional[builtins.int]=None, mesh_smoothing_iters:typing.Optional[builtins.int]=None, mesh_smoothing_weights:builtins.bool=True, mesh_smoothing_weights_normalization:builtins.float=13.0, generate_quads:builtins.bool=False, quad_max_edge_diag_ratio:builtins.float=1.75, quad_max_normal_angle:builtins.float=10.0, quad_max_interior_angle:builtins.float=135.0, output_mesh_smoothing_weights:builtins.bool=False, output_raw_normals:builtins.bool=False, output_raw_mesh:builtins.bool=False, mesh_aabb_min:typing.Optional[typing.Sequence[builtins.float]]=None, mesh_aabb_max:typing.Optional[typing.Sequence[builtins.float]]=None, mesh_aabb_clamp_vertices:builtins.bool=True) -> tuple[MeshWithData, SurfaceReconstruction]:
    r"""
    Runs the surface reconstruction pipeline for the given particle positions with optional post-processing
    
    Note that smoothing length and cube size are given in multiples of the particle radius.
    
    Parameters
    ----------
    particles : numpy.ndarray
        A two-dimensional numpy array of shape (N, 3) containing the positions of the particles.
    attributes_to_interpolate
        Dictionary containing all attributes to interpolate. The keys are the attribute names and the values are the corresponding 1D/2D arrays.
        The arrays must have the same length as the number of particles.
        Supported array types are 2D float32/float64 arrays for vector attributes and 1D uint64/float32/float64 arrays for scalar attributes.
    particle_radius
        Particle radius.
    rest_density
        Rest density of the fluid.
    smoothing_length
        Smoothing length of the SPH kernel in multiples of the particle radius (compact support radius of SPH kernel will be twice the smoothing length).
    cube_size
        Size of the cubes (voxels) used for the marching cubes grid in multiples of the particle radius.
    iso_surface_threshold
        Threshold of the SPH interpolation of the "color field" where the iso surface should be extracted.
    aabb_min
        Lower corner [x,y,z] of the AABB of particles that are active in the reconstruction.
    aabb_max
        Upper corner [x,y,z] of the AABB of particles to consider in the reconstruction.
    multi_threading
        Flag to enable multi-threading for the reconstruction and post-processing steps.
    simd
       Flag to enable SIMD vectorization for the reconstruction if supported by the CPU architecture.
    subdomain_grid
        Flag to enable spatial decomposition by dividing the domain into subdomains with dense marching cube grids for efficient multi-threading.
    subdomain_grid_auto_disable
        Flag to automatically disable the subdomain grid if the global domain is too small.
    subdomain_num_cubes_per_dim
        Number of marching cubes voxels along each coordinate axis in each subdomain if the subdomain grid is enabled.
    check_mesh_closed
        Flag to enable checking the final mesh for holes.
    check_mesh_manifold
        Flag to enable checking the final mesh for non-manifold edges and vertices.
    check_mesh_orientation
        Flag to enable checking the final mesh for inverted triangles (compares angle between vertex normals and adjacent face normals).
    check_mesh_debug
        Flag to enable additional debug output for the check-mesh operations (has no effect if no other check-mesh option is enabled).
    mesh_cleanup
        Flag to enable marching cubes mesh cleanup. This implements the method from "Compact isocontours from sampled data" (Moore, Warren; 1992).
    mesh_cleanup_snap_dist
        If marching cubes mesh cleanup is enabled, this limits vertex snapping to the specified distance relative to the MC edge length (should be in range of [0.0,0.5]).
    decimate_barnacles
        Flag to perform barnacle decimation. For details see "Weighted Laplacian Smoothing for Surface Reconstruction of Particle-based Fluids" (Löschner, Böttcher, Jeske, Bender; 2023).
    keep_vertices
        Flag to retain any vertices without connectivity resulting from mesh cleanup or decimation step instead of filtering them out.
    compute_normals
        Flag to enable computation of vertex normals on the final mesh.
    sph_normals
        Flag to enable computation of normals using SPH interpolation (gradient of the color field) instead of geometry-based normals.
    normals_smoothing_iters
        Number of Laplacian smoothing iterations to perform on the normal field.
    mesh_smoothing_iters
        Number of Laplacian smoothing iterations to perform on the mesh vertices.
    mesh_smoothing_weights
        Flag to enable computation and usage of mesh smoothing weights according to "Weighted Laplacian Smoothing for Surface Reconstruction of Particle-based Fluids" (Löschner, Böttcher, Jeske, Bender; 2023).
    mesh_smoothing_weights_normalization
        Normalization value for the mesh smoothing weights.
    generate_quads
        Flag to enable conversion of triangles to quads depending on geometric criteria.
    quad_max_edge_diag_ratio
        Maximum allowed ratio of quad edge lengths to its diagonals to merge two triangles to a quad (inverse is used for minimum).
    quad_max_normal_angle
        Maximum allowed angle (in degrees) between triangle normals to merge them to a quad.
    quad_max_interior_angle
        Maximum allowed vertex interior angle (in degrees) inside a quad to merge two triangles to a quad.
    output_mesh_smoothing_weights
        Flag to attach and return the smoothing weights as a mesh attribute if smoothing weights are computed.
    output_raw_normals
        Flag to output the raw normals in addition to smoothed normals if smoothing of normals is enabled.
    output_raw_mesh
        Flag to return a copy of the raw mesh before any post-processing steps (inside the returned reconstruction object).
    mesh_aabb_min
        Lower corner [x,y,z] of the axis-aligned bounding box for the mesh, triangles fully outside this box will be removed.
    mesh_aabb_max
        Upper corner [x,y,z] of the axis-aligned bounding box for the mesh, triangles fully outside this box will be removed.
    mesh_aabb_clamp_vertices
        Flag to clamp the vertices of the mesh to the AABB in addition to removing triangles outside the AABB.
    """



import pysplashsurf
import numpy as np
import meshio
import os.path
import pathlib
import tempfile

DIR = pathlib.Path(__file__).parent.resolve()
VTK_PATH = DIR.joinpath("ParticleData_Random_1000.vtk")


def test_aabb_class():
    print("\nTesting AABB class")

    aabb = pysplashsurf.Aabb3d.from_min_max(min=[0.0, 0.0, 0.0], max=[1.0, 2.0, 3.0])
    assert (aabb.min == np.array([0.0, 0.0, 0.0])).all()
    assert (aabb.max == np.array([1.0, 2.0, 3.0])).all()

    aabb = pysplashsurf.Aabb3d.from_min_max(
        min=np.array([0.0, 0.0, 0.0]), max=np.array([1.0, 2.0, 3.0])
    )
    assert (aabb.min == np.array([0.0, 0.0, 0.0])).all()
    assert (aabb.max == np.array([1.0, 2.0, 3.0])).all()

    aabb = pysplashsurf.Aabb3d.from_points(
        np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0], [2.0, 0.5, 4.2]])
    )

    print("AABB min:", aabb.min)
    print("AABB max:", aabb.max)

    assert (aabb.min == np.array([0.0, 0.0, 0.0])).all()
    assert (aabb.max == np.array([2.0, 1.0, 4.2])).all()

    assert aabb.contains_point([1.0, 0.9, 4.1])
    assert aabb.contains_point([0.0, 0.0, 0.0])
    assert not aabb.contains_point([2.0, 1.0, 4.2])
    assert not aabb.contains_point([1.0, -1.0, 5.0])


def impl_basic_test(dtype):
    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)

    mesh_with_data, reconstruction = pysplashsurf.reconstruction_pipeline(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0,
        cube_size=1.0,
        iso_surface_threshold=0.6,
        mesh_smoothing_iters=5,
        output_mesh_smoothing_weights=True,
    )

    assert type(mesh_with_data) is pysplashsurf.MeshWithData
    assert type(reconstruction) is pysplashsurf.SurfaceReconstruction
    assert type(mesh_with_data.mesh) is pysplashsurf.TriMesh3d

    mesh = mesh_with_data.mesh

    assert mesh_with_data.dtype == mesh.dtype
    assert mesh_with_data.dtype == dtype

    assert type(mesh_with_data.mesh_type) is pysplashsurf.MeshType
    assert mesh_with_data.mesh_type == pysplashsurf.MeshType.Tri3d

    assert mesh.vertices.dtype == dtype
    assert mesh.triangles.dtype in [np.uint32, np.uint64]

    assert mesh_with_data.nvertices == len(mesh.vertices)
    assert mesh_with_data.ncells == len(mesh.triangles)

    assert mesh_with_data.nvertices in range(21000, 25000)
    assert mesh_with_data.ncells in range(45000, 49000)

    assert mesh.vertices.shape == (mesh_with_data.nvertices, 3)
    assert mesh.triangles.shape == (mesh_with_data.ncells, 3)

    assert len(mesh_with_data.point_attributes) == 2
    assert len(mesh_with_data.cell_attributes) == 0

    assert "sw" in mesh_with_data.point_attributes
    assert "wnn" in mesh_with_data.point_attributes

    sw = mesh_with_data.point_attributes["sw"]
    wnn = mesh_with_data.point_attributes["wnn"]

    assert len(sw) == mesh_with_data.nvertices
    assert len(wnn) == mesh_with_data.nvertices

    assert sw.dtype == dtype
    assert wnn.dtype == dtype

    assert sw.shape == (mesh_with_data.nvertices,)
    assert wnn.shape == (mesh_with_data.nvertices,)

    assert sw.min() >= 0.0
    assert sw.max() <= 1.0

    assert wnn.min() >= 0.0


def test_pipeline_f32():
    impl_basic_test(np.float32)


def test_pipeline_f64():
    impl_basic_test(np.float64)


def reconstruct_test(dtype):
    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)

    reconstruction = pysplashsurf.reconstruct_surface(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0 * 0.025,
        cube_size=1.0 * 0.025,
        iso_surface_threshold=0.6,
        global_neighborhood_list=True,
    )

    assert type(reconstruction) is pysplashsurf.SurfaceReconstruction
    assert type(reconstruction.mesh) is pysplashsurf.TriMesh3d
    assert type(reconstruction.grid) is pysplashsurf.UniformGrid
    assert type(reconstruction.particle_densities) is np.ndarray
    assert type(reconstruction.particle_inside_aabb) is type(None)
    assert type(reconstruction.particle_neighbors) is pysplashsurf.NeighborhoodLists

    mesh = reconstruction.mesh

    assert mesh.dtype == dtype

    assert reconstruction.particle_densities.dtype == dtype
    assert len(reconstruction.particle_densities) == len(particles)

    assert len(mesh.vertices) in range(25000, 30000)
    assert len(mesh.triangles) in range(49000, 53000)


def test_reconstruct_f32():
    reconstruct_test(np.float32)


def test_reconstruct_f64():
    reconstruct_test(np.float64)


def neighborhood_search_test(dtype):
    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)

    reconstruction = pysplashsurf.reconstruct_surface(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0 * 0.025,
        cube_size=1.0 * 0.025,
        iso_surface_threshold=0.6,
        global_neighborhood_list=True,
    )

    neighbors_reconstruct = reconstruction.particle_neighbors.get_neighborhood_lists()

    assert type(neighbors_reconstruct) is list
    assert len(neighbors_reconstruct) == len(particles)

    aabb = reconstruction.grid.aabb

    neighbor_lists = pysplashsurf.neighborhood_search_spatial_hashing_parallel(
        particles, domain=aabb, search_radius=4.0 * 0.025
    )

    assert type(neighbor_lists) is pysplashsurf.NeighborhoodLists

    neighbors = neighbor_lists.get_neighborhood_lists()

    assert type(neighbors) is list
    assert len(neighbors) == len(particles)
    assert len(neighbors) == len(neighbors_reconstruct)

    # TODO: Compare with naive neighbor search


def test_neighborhood_search_f32():
    neighborhood_search_test(np.float32)


def test_neighborhood_search_f64():
    neighborhood_search_test(np.float64)


def check_consistency_test(dtype):
    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)

    reconstruction = pysplashsurf.reconstruct_surface(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0 * 0.025,
        cube_size=1.0 * 0.025,
        iso_surface_threshold=0.6,
        global_neighborhood_list=True,
    )
    mesh = reconstruction.mesh

    assert pysplashsurf.check_mesh_consistency(mesh, reconstruction.grid) is None

    mesh_with_data, reconstruction = pysplashsurf.reconstruction_pipeline(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0,
        cube_size=1.0,
        iso_surface_threshold=0.6,
        mesh_smoothing_iters=5,
        output_mesh_smoothing_weights=True,
    )

    assert (
        pysplashsurf.check_mesh_consistency(mesh_with_data, reconstruction.grid) is None
    )

    # TODO: Delete some triangles and check for failure


def test_check_consistency_f32():
    check_consistency_test(np.float32)


def test_check_consistency_f64():
    check_consistency_test(np.float64)


def tris_to_quads_test(dtype):
    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)

    mesh_with_data, reconstruction = pysplashsurf.reconstruction_pipeline(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0,
        cube_size=1.0,
        iso_surface_threshold=0.6,
        mesh_smoothing_iters=5,
        output_mesh_smoothing_weights=True,
    )

    mesh_with_data_quads = pysplashsurf.convert_tris_to_quads(mesh_with_data)

    assert type(mesh_with_data_quads.mesh) is pysplashsurf.MixedTriQuadMesh3d
    assert mesh_with_data_quads.mesh_type == pysplashsurf.MeshType.MixedTriQuad3d

    assert mesh_with_data_quads.nvertices == mesh_with_data.nvertices
    assert mesh_with_data_quads.ncells < mesh_with_data.ncells

    tris = mesh_with_data_quads.mesh.get_triangles()
    quads = mesh_with_data_quads.mesh.get_quads()

    assert tris.dtype in [np.uint32, np.uint64]
    assert quads.dtype in [np.uint32, np.uint64]

    assert len(tris) + len(quads) == mesh_with_data_quads.ncells

    assert tris.shape == (len(tris), 3)
    assert quads.shape == (len(quads), 4)

    assert len(tris) in range(35000, 39000)
    assert len(quads) in range(4600, 5000)

    assert len(mesh_with_data.point_attributes) == 2
    assert len(mesh_with_data.cell_attributes) == 0

    assert "sw" in mesh_with_data.point_attributes
    assert "wnn" in mesh_with_data.point_attributes


def test_tris_to_quads_f32():
    tris_to_quads_test(np.float32)


def test_tris_to_quads_f64():
    tris_to_quads_test(np.float64)


def interpolator_test(dtype):
    particles = np.array(meshio.read(VTK_PATH).points, dtype=dtype)

    mesh_with_data, reconstruction = pysplashsurf.reconstruction_pipeline(
        particles,
        particle_radius=0.025,
        rest_density=1000.0,
        smoothing_length=2.0,
        cube_size=1.0,
        iso_surface_threshold=0.6,
        mesh_smoothing_iters=5,
        output_mesh_smoothing_weights=True,
    )

    compact_support = 4.0 * 0.025
    rest_mass = 1000.0 * 0.025**3

    interpolator = pysplashsurf.SphInterpolator(
        particles, reconstruction.particle_densities, rest_mass, compact_support
    )

    assert type(interpolator) is pysplashsurf.SphInterpolator

    mesh = mesh_with_data.mesh
    mesh_densities = interpolator.interpolate_quantity(
        reconstruction.particle_densities, mesh.vertices
    )

    assert type(mesh_densities) is np.ndarray
    assert mesh_densities.dtype == dtype
    assert mesh_densities.shape == (len(mesh.vertices),)
    assert mesh_densities.min() >= 0.0

    mesh_particles = interpolator.interpolate_quantity(particles, mesh.vertices)

    assert type(mesh_particles) is np.ndarray
    assert mesh_particles.dtype == dtype
    assert mesh_particles.shape == (len(mesh.vertices), 3)

    mesh_sph_normals = interpolator.interpolate_normals(mesh.vertices)

    assert type(mesh_sph_normals) is np.ndarray
    assert mesh_sph_normals.dtype == dtype
    assert mesh_sph_normals.shape == (len(mesh.vertices), 3)

    mesh_with_data.add_point_attribute("density", mesh_densities)
    mesh_with_data.add_point_attribute("position", mesh_particles)
    mesh_with_data.add_point_attribute("normal", mesh_sph_normals)

    assert "density" in mesh_with_data.point_attributes
    assert "position" in mesh_with_data.point_attributes
    assert "normal" in mesh_with_data.point_attributes

    assert np.array_equal(mesh_with_data.point_attributes["density"], mesh_densities)
    assert np.array_equal(mesh_with_data.point_attributes["position"], mesh_particles)
    assert np.array_equal(mesh_with_data.point_attributes["normal"], mesh_sph_normals)


def test_interpolator_f32():
    interpolator_test(np.float32)


def test_interpolator_f64():
    interpolator_test(np.float64)
